/*
Copyright (c) 2009, Eamon O'Dea <odea35 [at] gmail.com>

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
/*
 * command to compile:
 * cc -g -Wall -I\. -lm -lgsl -lgslcblas -o ExtinCalc ExtinCalc.c mtype_branching_sims.c file_parser.c
 */

#include <stdio.h>
#include <math.h>
#include <sim.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_randist.h>
#include <float.h>

#define TOLERANCE FLT_EPSILON

void
makeBins (int l, double *alpha, double *beta,
	  gsl_matrix_short * bins, short numbins);
int
fixedPointSolver (unsigned long l, short numbins, 
		  double *vec, int b, double sv, int pois, int mkdot,
                  int do_sim, int mean_mat);
int
fixedPtIteration (double *vec, short numbins, unsigned long l,
		  double q, double *alphas, double *betas,
		  double *row, gsl_matrix_short * bins, double sv, int b);
void getBinPMF (double *binPMF, unsigned long l, double q);
void
parseCL (int argc, char *arcv[], unsigned long *l, 
         short *numbins, int *b, double *sv, int *pois, int *mkdot,
         int *do_sim, int *mean_mat);

int
makeDotFile (gsl_matrix_short * bins, unsigned int l,
	     short numbins, double *binPMF, double q,
	     double *betas, int b, double sv, double *alphas);

int fixedPtItPois (double *vec, short numbins, unsigned long l,
		   double q, double *alphas, double *betas,
		   double *binPMF, gsl_matrix_short * bins, double sv, int b);

int file_parser (double *alphas, double *betas, short numbins);

int
get_mean_matrix (short numbins, double *alphas, double *betas,
                 double *binPMF, gsl_matrix_short * bins, int l,
                 double sv, int b, gsl_matrix * mean_matrix);

int
main (int argc, char *argv[])
{
  int ret;
  /* value to return, 0 means no errors */

  unsigned long int l;
  /* sequnce length */
  
  short numbins;                
  /* number of bins to split selection coefficient into */

  int b;                        
  /* burst size parameter */
  
  int pois;
  /* 1=use Poisson offspring distribution */

  int mkdot;
  /* 0=don't write file for graphviz, 1 = do it */

  int do_sim;
  /* 0=don't do stochastic branching simulation, 
   * 0<do_sim < 100000 do DO_SIM reps  */
  
  int mean_mat;
  /* 0=don't make mean matrix file, 1 = do it, 2 = do it AND
   * don't bother calculating extincion probs */

  double sv;
  /* survival probability */


  double *vec;
  /* vector of extinction probabilities */

  parseCL (argc, argv, &l, &numbins, &b, &sv, &pois, &mkdot, &do_sim, &mean_mat);

  vec = (double *) malloc (numbins * sizeof (double));
  if (!vec)
    {
      printf ("%s: %d: malloc of vec array failed\n", __FILE__, __LINE__ );
      return (1);
    }
  ret = fixedPointSolver (l, numbins, vec, b, sv, pois, mkdot, do_sim, mean_mat);

  if (vec)
    free (vec);

  return (ret);
}



void
makeBins (int l, double *alphas, double *betas, gsl_matrix_short * bins,
	  short numbins)
{
  short i, binNum;
  unsigned long int n;
  double x;
  double OneLessS;		/* 1 - s, where s is fraction of
				   deleterious mutations at one step */
  double s;

  for (i = 0; i < numbins - 1; i++)
    {
      gsl_matrix_short_set (bins, i, 0, i);	/*from definition, avoids rounding error */
      for (n = 1; n <= l; n++)
	{
	  x = pow (n, betas[i]) - pow (n + 1, betas[i]);
	  OneLessS = exp (alphas[i] * x);	/*  = w(n+1) / w(n) */
	  s = 1 - OneLessS;
          if (s < 1){
              /* cast as int to get bin # */
              binNum = (short) (s * numbins);
          }
          else if (s == 1){
              /* Avoid rounding error putting offpring in nonexistent bin */
              binNum = numbins - 1;
          }
          else{
              fprintf (stderr, "Error: %s: %d: s > 1\n", __FILE__, __LINE__);
          }
	  gsl_matrix_short_set (bins, i, n, binNum);		}
    }
  i = numbins - 1;
  gsl_matrix_short_set (bins, i, 0, i);
  /*In the highest bin, only offspring with zero mutations can survive, so the BINS 
   * element just set should be the only one in this row ever used. Next we fill in
   * the rest of the row with a value that are likely to cause an error if they are
   * ever requested by accident.
   */
  for (n = 1; n <=l; n++)
  {
	  gsl_matrix_short_set (bins, i, n, -99);
  }

}

int
fixedPointSolver (unsigned long l, short numbins, 
		  double *vec, int b, double sv, int pois, int mkdot,
                  int do_sim, int mean_mat)
{
  short i;
  int ret;
  double q;
  char string[100];
  double *alphas, *betas, *binPMF = NULL;

  gsl_matrix_short *bins;
  /*matrix where the jth column of the ith row represents the 
   * bin number of the offspring of a type i parent that has offspring
   * with j mutations*/

  gsl_matrix *mat;
  /*matrix of means*/

  FILE *ab_scatter = NULL, *extn_surface = NULL, *matrix_output=NULL;

  alphas = (double *) malloc ((numbins) * sizeof (double));
  if (!alphas)
    {
      fprintf (stderr, "Error: %s: %d: Malloc of alphas array failed\n",
	       __FILE__, __LINE__);
      return (1);
    }

  betas = (double *) malloc ((numbins-1) * sizeof (double));
  /* the highest bin has fitness of 0 with any mutations, so we don't need that beta, thus numbins -1 */
  if (!betas)
    {
      fprintf (stderr, "Error: %s: %d: Malloc of betas array failed\n",
	       __FILE__, __LINE__);
      return (1);
    }
  file_parser (alphas, betas, numbins);
  if ((ab_scatter = fopen ("ab_scatter", "w")) == NULL)
    {
      fprintf (stderr, "Error: %s: %d: Failed to open file ab_scatter\n",
	       __FILE__, __LINE__);
      return (-1);
    }
  fprintf (ab_scatter, "alpha, beta\n");
  for (i = 0; i < numbins - 1; i++){
      fprintf (ab_scatter, "%g, %g\n", alphas[i], betas[i]);
/*      printf ("\n%g, %g\n", alphas[i], betas[i]);*/
  }
  bins = gsl_matrix_short_alloc (numbins, l + 1);
  makeBins (l, alphas, betas, bins, numbins);
/*  gsl_matrix_short_fprintf (stdout, bins, "%d"); */

  if ((extn_surface = fopen ("extn_surface", "w")) == NULL)
    {
      fprintf (stderr, "cannot open file extn_surface \n");
      return (-1);
    }

  binPMF = (double *) malloc ((l + 1) * sizeof (double));
  if (!binPMF)
    {
      fprintf (stderr, "Errof: %s: %d: Malloc of binPMF array failed\n",
	       __FILE__, __LINE__);
      return (1);
    }
  


  double qmin = 1 - (double) 20 / l;
  double qinc = (1 - qmin) / 100;
  fprintf (stdout, "Error tolerance for fixed point iteration: %g\n", (double) TOLERANCE );
  fprintf (stdout, "\n");
  fprintf (stdout, "Progress\n");
  fprintf (stdout, "--------\n");
  fprintf (stdout, "Extinction probabilities calculated for these fidelity rates q:\n   ");
  int newline_fuse = 10; 
  /* Use newline_fuse to put newlines in progress printing */
  if (!pois)
    {
      for (q = 1 - (double) 20 / l; q < 1; q = q + qinc)
	{
	  getBinPMF (binPMF, l, q);
	  if (mkdot)
	    ret =
	      makeDotFile (bins, l, numbins, binPMF, q, betas, b, sv, alphas);
          if (mean_mat != 2){
              fixedPtIteration (vec, numbins, l, q, alphas, betas, binPMF, bins,
                                sv, b);
              fprintf (stdout, "%3.3f  ", q);
          }
          if (mean_mat)
            {
              sprintf (string, "mean_mat_qis0p%6.0f.dat", 1000000 * q);
              if ((matrix_output = fopen (string, "w")) == NULL)
                {
                  fprintf (stderr, "%s: %d: Cannot open file for writing\n",
                           __FILE__, __LINE__);
                  return (-1);
                }
              mat = gsl_matrix_alloc (numbins, numbins);
              get_mean_matrix (numbins, alphas, betas,binPMF, 
                               bins, l, sv, b, mat);
              gsl_matrix_fprintf (matrix_output, mat, "%g");
              gsl_matrix_free(mat);
              if (matrix_output)
                {
                  fclose (matrix_output);
                }
              
            }
	  if (--newline_fuse <= 0)
            {
		  fprintf(stdout, "\n   ");
		  newline_fuse = 10;
            }

	  for (i = 0; i < numbins - 1; i++)
	    fprintf (extn_surface, "%g,", vec[i]);
	  fprintf (extn_surface, "%g\n", vec[i]);
	}
    }
  else
    {
      for (q = 1 - (double) 20 / l; q < 1; q = q + qinc)
	{
	  getBinPMF (binPMF, l, q);
	  if (mkdot)
            {
              ret =
                makeDotFile (bins, l, numbins, binPMF, q, betas, b, sv, alphas);
            }
          if (mean_mat != 2){
              fixedPtItPois (vec, numbins, l, q, alphas, betas, binPMF, bins, sv,
                             b);
              fprintf (stdout, "%3.3f  ", q);
          }
          if (mean_mat)
            {
              sprintf (string, "mean_mat_qis0p%6.0f.dat", 1000000 * q);
              if ((matrix_output = fopen (string, "w")) == NULL)
                {
                  fprintf (stderr, "%s: %d: Cannot open file for writing\n",
                           __FILE__, __LINE__);
                  return (-1);
                }
              mat = gsl_matrix_alloc (numbins, numbins);
              get_mean_matrix (numbins, alphas, betas,binPMF, 
                               bins, l, sv, b, mat);
              gsl_matrix_fprintf (matrix_output, mat, "%g");
              gsl_matrix_free(mat);
              if (matrix_output)
                {
                  fclose (matrix_output);
                }
            }
          if (--newline_fuse <= 0)
            {
              fprintf(stdout, "\n   ");
              newline_fuse = 10;
            }
	  for (i = 0; i < numbins - 1; i++)
            {
              fprintf (extn_surface, "%g,", vec[i]);
            }
          fprintf (extn_surface, "%g\n", vec[i]);
	}
    }
  printf ("\n");
  if (do_sim)
    {
      if (pois)
        {
          sim_stepper(do_sim, l, b, sv, numbins, pois, bins, alphas, betas);
        }
      else
        {
          fprintf (stderr, "Warning: %s: %d: Only process with Poisson distribution of offspring can be simulated at this time\n", __FILE__, __LINE__);
        }
    }

  if (extn_surface)
    fclose (extn_surface);
  if (ab_scatter)
    fclose (ab_scatter);
  if (alphas)
    free (alphas);
  if (betas)
    free (betas);
  if (binPMF)
    free (binPMF);
  if (bins)
    gsl_matrix_short_free (bins);
  return 0;
}

void
getBinPMF (double *binPMF, unsigned long l, double q)
{
/*
 * make a table for the distribution of the number of mutations
 * since "l" is large and "1-q" is small, use Poisson approximation
 * */
  int n;
  double mu;
/*
 *   if (!(l >= 20 && 1 - q <= 0.05))
 *      fprintf (stdout, "use of Poisson approximation to the binomial/
 * may not be to good for these parameters\n");
 */
    mu = l * (1 - q);
  for (n = 0; n <= l; n++)
    binPMF[n] = gsl_ran_poisson_pdf (n, mu);
}

void
parseCL (int argc, char *argv[], unsigned long *l, 
         short *numbins, int *b, double *sv, int *pois,
         int *mkdot, int *do_sim, int *mean_mat)
{
  if (argc != 17)
    {
      printf
	("usage: ExtinCalc -l <unsigned long l> -numbins <short> -b <int> -sv <double> -pois <1>or<0> -makedot <1>or<0> -simulate <1>or<0> -mean_mat <1> or <0>\n");
      printf( "\n%d\n", argc);
      exit (1);
    }

  fprintf (stdout, "\nInputs\n");
  fprintf (stdout, "------\n");

  *l = atoll (argv[2]);
  if (*l < 1 || *l > 1000000)
    {
      printf ("problem with l\n");
      exit (1);
    }
  fprintf (stdout, "sequence length = %ld\n", *l);

  *numbins = (short) atoi (argv[4]);
  if (*numbins < 1 || *numbins > 1024)
    {
      printf ("problem with numbins\n");
      exit (1);
    }
  fprintf (stdout, "number of bins for s = %d\n", *numbins);

  *b = atoi (argv[6]);
  if (*b < 1 || *b > 1000000)
    {
      printf ("problem with b\n");
      exit (1);
    }
  fprintf (stdout, "burst size = %d\n", *b);

  *sv = (double) atof (argv[8]);
  if (*sv < 0 || *sv > 1)
    {
      printf ("problem with sv\n");
      exit (1);
    }
  fprintf (stdout, "survival probability = %g\n", *sv);

  *pois = atoi (argv[10]);
  if (*pois != 0 && *pois != 1)
    {
      printf ("problem with pois\n");
      exit (1);
    }
  if (*pois == 0)
    {
      fprintf (stdout, "Offpspring distribution is binomial\n");
    }
  if (*pois == 1)
    {
      fprintf (stdout, "Offpspring distribution is Poisson\n");
    }


  *mkdot = atoi (argv[12]);
  if (*mkdot != 0 && *mkdot != 1)
    {
      printf ("problem with mkdot argument\n");
      exit (1);
    }
  if (*mkdot == 0)
    {
      fprintf (stdout, "Not making dot file\n");
    }
  if (*mkdot  == 1)
    {
      fprintf (stdout, "Making dot file\n");
    }

  *do_sim = atoi (argv[14]);
  if (*do_sim < 0 || *do_sim > 100000)
    {
      printf ("problem with do_sim argument\n");
      exit (1);
    }
  fprintf (stdout, "Doing %d replicates of simulations\n", *do_sim);

  *mean_mat = atoi (argv[16]);
  if (*mean_mat != 0 && *mean_mat != 1 && *mean_mat != 2)
    {
      printf ("problem with mean_mat argument\n");
      exit (1);
    }
  if (*mean_mat == 0)
    {
      fprintf (stdout, "Not making file of mean matrices\n");
    }
  if (*mean_mat  == 1)
    {
      fprintf (stdout, "Making file of mean matrices\n");
    }
  if (*mean_mat  == 2)
    {
      fprintf (stdout, "Making file of mean matrices");
      fprintf (stdout, "AND not calculating extinction probabilities\n");
    }
  fprintf (stdout, "\nNo problems found with inputs\n\n");
}

int
makeDotFile (gsl_matrix_short * bins, unsigned int l,
	     short numbins, double *binPMF, double q,
	     double *betas, int b, double sv, double *alphas)
{
  int i, n;
  double weight, source, sink, lastSink, r;
  char string[100];
  FILE *fp;

  /*Makes graphiz dot file for drawing directed networks based on
   * the offpring from one type to another.
   */

  r = b * sv;			/*only draw edges when there's expected to be at least one offpring there */

  sprintf (string, "network_qis0p%6.0f.gv", 1000000 * q);
  if ((fp = fopen (string, "w")) == NULL)
    {
      fprintf (stderr, "cannot open file makeDotFile() \n");
      return (-1);
    }
  fprintf (fp, "digraph G {\n\tsize=\"4,8\";\n\tratio= compact;\n");
  fprintf (fp, "\tnode [style = filled];\n");
/*  uncomment this to line up the nodes */
/*  fprintf(fp,"subgraph cluster1 { \n edge [color=white, arrowsize = 0];\n");
  for (i=0; i<numbins-1; i++)
    fprintf(fp,"%g->",((double) (i+1))/numbins);
  fprintf(fp,"%g;\n}\n", ((double) (i+1))/numbins); 
  fprintf(fp, "edge [color=black]\n");
*/
  for (i = 0; i < numbins - 1; i++)
    {
      source = ((double) (i + 1)) / numbins;
      weight = 0;
      lastSink = ((double) (gsl_matrix_short_get (bins, i, 0) + 1)) / numbins;
      for (n = 0; n <= l; n++)
	{
	  sink = ((double) (gsl_matrix_short_get (bins, i, n) + 1)) / numbins;

	  if (sink == lastSink)
	    {
	      weight += binPMF[n] * exp (-1 * alphas[i] * pow (n, betas[i]));
/*              if (betas[i] < 0.99 && q > 0.99 )
                if (n < 14)
                  printf("source %f, sink %f, n = %d, binsPMF=%f, w=%g, weight = %f, alphas[i]=%f, betas[i]=%f\n",source, sink,  n, binPMF[n], exp(-1*alphas[i]*pow(n,betas[i])),weight, alphas[i], betas[i]);*/

	    }
	  else
	    {
	      if (weight >= 1 / r)
		fprintf (fp, "%g -> %g;\n", source, lastSink);
	      weight = binPMF[n];
	      lastSink = sink;
	    }
	  if (sink == lastSink && n == l)
	    {
	      if (weight >= 1 / r)
		fprintf (fp, "%g -> %g;\n", source, lastSink);
	      weight = binPMF[n];
	      lastSink = sink;
	    }
	}
      if (betas[i] > 1.01)
	fprintf (fp, "%g [fillcolor=green];\n", source);
      else if (betas[i] < 0.99)
	fprintf (fp, "%g [fillcolor=orange];\n", source);
    }
  if (binPMF[0] >= 1 / r)
    fprintf (fp, "1 -> 1;\n");
  fprintf (fp, "1 [fillcolor=orange];\n}");
  if (fp)
    fclose (fp);
  return 0;
}

int
fixedPtIteration (double *vec, short numbins, unsigned long l,
		  double q, double *alphas, double *betas,
		  double *binPMF, gsl_matrix_short * bins, double sv, int b)
{
  /* iterates pgf's corresponding to multinomial offspring distribution
   * in order to get probability of extinction */

  int iter = 0, maxIter = 2000;
  short i;
  unsigned long n;
  double accum, reduce, term, diff, *vecNew, *temp;

  vecNew = (double *) malloc (numbins * sizeof (double));
  if (!vecNew)
    {
      printf ("malloc of vecNew array in fixedPtIteration() failed");
      return (1);
    }

  /* initialize the vectors of extinction probabilities with a guess */
  for (i = 0; i < numbins; i++)
    vecNew[i] = 0.00001;

  do
    {
      iter++;
      temp = vec;
      vec = vecNew;
      vecNew = temp;
      for (i = 0; i < numbins - 1; i++)
	{
	  accum = 0;
	  reduce = 0;
	  for (n = 0; n <= l; n++)
	    {
	      term = binPMF[n];
	      term *= exp (-alphas[i] * pow (n, betas[i]));
	      accum += term * vec[gsl_matrix_short_get (bins, i, n)];
	      reduce -= term;
	    }
	  accum *= sv;
	  reduce *= sv;
	  if ((accum + reduce + 1) > 1)
	    {
	      printf ("error:  %f > 1 in pgf i=%d \n", accum + 1 + reduce, i);
	      return -1;
	    }
	  vecNew[i] = pow (accum + (1 + reduce), b);
	}
      /* special case of vecNew[numbins -1 ]: w(0) = 1, all other w = 0 */
      vecNew[numbins - 1] =
	pow (binPMF[0] * sv * vec[numbins - 1] + (1 - sv * binPMF[0]), b);
      diff = 0;
      for (i = 0; i < numbins; i++)
	diff += fabs (vecNew[i] - vec[i]);
/*      printf("diff= %f\n",diff);*/
    }
  while (diff > TOLERANCE && iter < maxIter);
  if (diff <= TOLERANCE)
    return 0;
  fprintf (stderr, "maximum number of iterations=%d reached, q=%g\n", maxIter,
	   q);
  return -1;
}



int
fixedPtItPois (double *vec, short numbins, unsigned long l,
	       double q, double *alphas, double *betas,
	       double *binPMF, gsl_matrix_short * bins, double sv, int b)
{
  /* iterates pgf's corresponding to multivariate poisson
   *  offspring distribution in order to get probability of 
   *  extinction */

  int iter = 0, maxIter = 10000;
  short i;
  unsigned long n;
  double accum, reduce, term, diff, *vecNew, *temp;
  double net_offspring;

  net_offspring = b * sv;

  vecNew = (double *) malloc (numbins * sizeof (double));
  if (!vecNew)
    {
      printf ("malloc of vecNew array in fixedPtItPois() failed");
      return (1);
    }

  /* initialize the vectors of extinction probabilities with a guess */
  for (i = 0; i < numbins; i++)
    vecNew[i] = 0.00001;

  do
    {
      iter++;
      temp = vec;
      vec = vecNew;
      vecNew = temp;
      for (i = 0; i < numbins - 1; i++)
	{
	  accum = 0;
	  reduce = 0;
	  for (n = 0; n <= l; n++)
	    {
	      term = binPMF[n];
	      term *= exp (-alphas[i] * pow (n, betas[i]));
	      accum += term * vec[gsl_matrix_short_get (bins, i, n)];
	      reduce -= term;
	    }
	  if (accum + reduce > 0 || accum + reduce < -1.0)
	    {
	      fprintf (stderr, "warning:  in poisson pgf i=%d, accum + reduce = %g \n",
                       i, accum + reduce);
	     /* return -1;*/
	    }
	  vecNew[i] = exp (net_offspring * (accum + reduce));
	}
      /* special case of vecNew[numbins -1 ]: w(0) = 1, all other w = 0 */
      vecNew[numbins - 1] =
	exp (net_offspring * (binPMF[0] * vec[numbins - 1] - binPMF[0]));
      diff = 0;
      for (i = 0; i < numbins; i++)
	diff += fabs (vecNew[i] - vec[i]);
    }
  while (diff > TOLERANCE && iter < maxIter);
  if (diff <= TOLERANCE)
    return 0;
  fprintf (stderr,
	   "Warning: %s: %d: Maximum number of iterations=%d reached\n",
	   __FILE__, __LINE__, maxIter);
  fprintf (stderr, "\tdifference=%g > tolerance=%g, fidelity q=%g\n", diff,
	   TOLERANCE, q);
  return -1;
}



int
get_mean_matrix (short numbins, double *alphas, double *betas,
                 double *binPMF, gsl_matrix_short * bins, int l,
                 double sv, int b, gsl_matrix * mean_matrix)
{
  /* returns the matrix of the expected number of
   *  offspring of any one type I to any one of the possible types R
   */

  size_t i,j;
  size_t offspring_bin_number;
  size_t n;
  double old_value, new_value, term;
  double net_offspring;

  net_offspring = b * sv;

  for (i=0; i < numbins -1; i++)
    {
      for (j = 0; j < numbins - 1; j++)
        {
          gsl_matrix_set (mean_matrix, i, j, 0);
        }
    }


  for (i = 0; i < numbins - 1; i++)
    {
      for (n = 0; n <= l; n++)
	    {
	      term = binPMF[n];
	      term *= exp (-alphas[i] * pow (n, betas[i]));
              if (term > 1){
                  fprintf(stderr, "term that should be <= 1 is %g", term);
              }
	      offspring_bin_number = gsl_matrix_short_get (bins, i, n);
              old_value = gsl_matrix_get (mean_matrix, i, offspring_bin_number);
              if (old_value > net_offspring){
                  fprintf(stderr, "old_value that should be < burst size is %g", old_value);
              }
              new_value = old_value + term;
              gsl_matrix_set (mean_matrix, i, offspring_bin_number, 
                              new_value);

	    }
    }
  gsl_matrix_scale (mean_matrix, net_offspring);
/*
  for (i=0; i < numbins -1; i++)
    {
      for (j = 0; j < numbins - 1; j++)
        {
          fprintf (stdout, "[%d,%d]=%g ", i,j, gsl_matrix_get(mean_matrix, i, j) );
        }
      fprintf (stdout, "\n");
    }
*/
  return 0;
}
