/*
Copyright (c) 2009, Eamon O'Dea <odea35 [at] gmail.com>

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
#include <sim.h>

int
sim_stepper (int num_reps, int l, int b, double sv,
	     int numbins, int pois, gsl_matrix_short * bins,
	     double *alphas, double *betas)
{
  /*steps through parameters for simulation and writes results to 
   * file*/

  int rep, num_extinctions, ret, bin, extinct;
  int num_mus = 100;
  FILE *data_table;
  double p_extn, u, u_max, u_inc;

  int
    (*multi_branching_sim) (int *extinct, int bin, int numbins,
			    double u, int l, int b, double sv,
			    gsl_matrix_short * bins, double *alphas,
			    double *betas, gsl_rng * r);
  if (pois)
    multi_branching_sim = multi_branching_sim_pois;
  else
    multi_branching_sim = multi_branching_sim_bin;

  if ((data_table = fopen ("simulation_results.csv", "w")) == NULL)
    {
      fprintf (stderr, "failed to open data output file in\
               simulations()\n");
      return 1;
    }
  fprintf (data_table, "reps, p_extn, bin, mu, l, b, sv, numbins,\
           betas\n");

  /* Initialize the taus2 random number generator in the 
   * Gnu Scientific Lib.*/
  gsl_rng *r = gsl_rng_alloc (gsl_rng_taus2);

  u_max = 20;
  u_inc = u_max / num_mus;
  for (u = u_max; u >= u_inc; u = u - u_inc)
    for (bin = 0; bin < numbins; bin++)
      {
	num_extinctions = 0;
	for (rep = 0; rep < num_reps; rep++)
	  {
	    ret += multi_branching_sim (&extinct, bin, numbins,
					u, l, b, sv, bins, alphas, betas, r);
	    num_extinctions += extinct;
	  }
	p_extn = (double) num_extinctions / num_reps;
	if (bin < numbins -1)
	{
		fprintf (data_table, "%d, %g, %d, %g, %d, %d, %g, %d,\
%g\n", num_reps, p_extn, bin, u, l, b, sv, numbins, betas[bin]);
	}
	else
	{
		fprintf (data_table, "%d, %g, %d, %g, %d, %d, %g, %d,\
NA\n", num_reps, p_extn, bin, u, l, b, sv, numbins);

	}

      }
  return ret;
}

int
multi_branching_sim_pois (int *extinct, int bin, int numbins,
			  double u, int l, int b, double sv,
			  gsl_matrix_short * bins, double *alphas,
			  double *betas, gsl_rng * r)
{
  /* This function simulates a multitype branching process.  Each
   * individual has a Poisson distributed number of offspring,
   * each offspring then has a Poisson distributed number of 
   * mutations, and the offspring is neutral depending on the function
   * \exp (-\alpha * mutations^beta).  The type of that offspring
   * determines it's alpha and beta. <int bin> is the type of the 
   * founding individual. *bins is the a matrix where the rows $i$
   * correspond to types and the columns $j$ correspond to the type 
   * number of the offspring after $j$ mutations.*/

  int i, j, c, *xc, *xn, num_off, num_muts, total, last_total;
  int generation, *temp;
  int max_gen = 10000;
  int explosion_cutpoint = 200;
  double mean_off, prob_neut, rand;

  /*Allocate memory for arrays the hold the population sizes 
   * for each type in the current generation, xc, and the next,
   * xn. */
  xc = (int *) malloc (numbins * sizeof (int));
  xn = (int *) malloc (numbins * sizeof (int));
  if (!xc || !xn)
    {
      printf ("malloc of array in multi_branching_sim() failed");
      return (1);
    }
  mean_off = b * sv;

  /* Start with one individual of type BIN and zero individuals of all other types. */
  for (i = 0; i < numbins; i++)
    {
      xn[i] = 0;
      xc[i] = 0;
    }
  xc[bin] = 1;
  last_total = 1;
  generation = 0;
  do
    {
	    /* For each level of neutrality, */
      for (i = 0; i < numbins; i++)
	{
		/* for each individual, */
	  for (c = 1; c <= xc[i]; c++)
	    {
		    /* get the random number of offspring they have. */
	      num_off = gsl_ran_poisson (r, mean_off);
	      /* For each offspring, */
	      for (j = 1; j <= num_off; j++)
		{
			/* give it a random number of mutations that is less than
			 *  the sequence length. */
		  do
		    num_muts = gsl_ran_poisson (r, u);
		  while (num_muts > l);
		  /* Now calculate the probability that the offspring 
		   * with NUM_MUTS mutations will be neutral. The highest
		   * bin, indexed with NUMBINS - 1,  must be treated as a 
		   * special case because we have defined it so that 
		   * all mutations will be non-neutral. */

                  if (i != numbins - 1)
                    {
                      prob_neut = exp (-alphas[i] * pow (num_muts, betas[i]));
                    }
                  else if (i == numbins - 1 && num_muts > 0)
                    {
                      prob_neut = 0;
                    }
                  else 
                    {
                      prob_neut = 1;
                    }
		  /*Now get a random variate and determine the neutrality of 
		   * the offspring with the above probability. */
		  rand = gsl_rng_uniform (r);
//                  fprintf (stderr, "generation =%d, total = %d, bin = %d, muts = %d, pneut=%f, rand = %f, num_off = %d, c = %d\n", generation, total, bin, num_muts, prob_neut, rand, num_off, c);
/* 
 * If the offspring are neutral, add then to the array that tracks the number of
 * individuals of each level of neutrality in the next generation.
 */
		  if (rand < prob_neut)
                    {
//                      printf("%d, %d\n", i, bin);
                      xn[gsl_matrix_short_get (bins, i, num_muts)]++;
                    }
		}
	    }
	  /*
	   * After we have each the contribution of each individual in this generation
	   * in this bin to the next generation, we zero out the population. This
	   * may seem unncessariy but it allows us to use this area of memory for the 
	   * tally of the next generation (the XN array) in the next generation by just
	   * swapping the pointers.
	   */
	  xc[i] = 0;		/* Individuals only live one generation */
	}
      /*
       * Now we have gone through all the levels of neutrality. So we can get the total
       * size of the next generation.
       */
      generation++;
      total = 0;
      for (i = 0; i < numbins; i++)
	total += xn[i];
      if (total == 0)
	{
	  *extinct = 1;
	  free (xn);
	  free (xc);
	  return 0;
	}
      else
	{
	  last_total = total;
	  temp = xn;
	  xn = xc;		/*This should be an array of zeroes. */
	  xc = temp;
	}
    }
  while (generation < max_gen && total < explosion_cutpoint);
  if (total >= explosion_cutpoint)
    {
      *extinct = 0;
      free (xn);
      free (xc);
      return 0;
    }
  else
    {
      fprintf (stderr, "%s:%d: \
               error: Exceeded maximum number of generations\n", __FILE__, __LINE__);
      free (xn);
      free (xc);
      return 1;
    }
}

int
multi_branching_sim_bin (int *extinct, int bin, int numbins,
			 double mu, int l, int b, double sv,
			 gsl_matrix_short * bins, double *alphas,
			 double *betas, gsl_rng * r)
{

  *extinct = 1;
  return 0;
}
